/*********************************************************************
; Copyright 1992-97, The University of Iowa (UI).  All rights reserved. 
; By using this software the USER indicates that he or she has read, 
; understood and will comply with the following:
;
; --- UI hereby grants USER nonexclusive permission to use, copy and/or
; modify this software for internal, noncommercial, research purposes only.
; Any distribution, including commercial sale or license, of this software,
; copies of the software, its associated documentation and/or modifications
; of either is strictly prohibited without the prior consent of UI.  Title
; to copyright to this software and its associated documentation shall at
; all times remain with UI.  Appropriate copyright notice shall be placed
; on all software copies, and a complete copy of this notice shall be
; included in all copies of the associated documentation.  No right is
; granted to use in advertising, publicity or otherwise any trademark,
; service mark, or the name of UI.  Software and/or its associated
; documentation identified as "confidential," if any, will be protected
; from unauthorized use/disclosure with the same degree of care USER
; regularly employs to safeguard its own such information.
;
; --- This software and any associated documentation is provided "as is," and
; UI MAKES NO REPRESENTATIONS OR WARRANTIES, EXPRESS OR IMPLIED, INCLUDING
; THOSE OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE, OR THAT
; USE OF THE SOFTWARE, MODIFICATIONS, OR ASSOCIATED DOCUMENTATION WILL
; NOT INFRINGE ANY PATENTS, COPYRIGHTS, TRADEMARKS OR OTHER INTELLECTUAL
; PROPERTY RIGHTS OF A THIRD PARTY.  UI, the University of Iowa,
; its Regents, officers, and employees shall not be liable under any
; circumstances for any direct, indirect, special, incidental, or
; consequential damages with respect to any claim by USER or any third
; party on account of or arising from the use, or inability to use, this
; software or its associated documentation, even if UI has been advised
; of the possibility of those damages.
*********************************************************************/
#include "main.h"

#define TRIPLE2KEY(i,j,k) (((i) * n_sqt) + ((j) * n) + (k) + 1)
#define OMEGA(i,j,k) (((i) * n_sqt) + ((j) * n) + (k) + 1 + n_cube)
#define PHI(i,j,k) (((i) * n_sqt) + ((j) * n) + (k) + 1 + n_cube)
#define P2(i,j,k,l) ((((i) * n_sqt) + ((j) * n) + k) * n + l + 1 + n_cube)
#define P3(i,j,k,l) ((((i) * n_sqt) + ((j) * n) + k) * n + l + n_four)
#define neg(x) (n1-x)

#define special_call(x,y,z,w,u,v) insert_cl_4(TRIPLE2KEY(x,y,u),TRIPLE2KEY(z,w,u),TRIPLE2KEY(y,x,v),TRIPLE2KEY(w,z,v),0,cl_arr,sign_arr)

int num_of_hole = 0;
int double_cl = 0;
int m = 0;
int n = 0;
int n1 = 0;
int n_sqt, n_cube, n_four;

void locate_holes();

int get_num_of_hole () { return num_of_hole; }
void set_num_of_hole (x) int x; { num_of_hole = x; }

int multi_squares () { return double_cl; }
void set_double_cl (x) int x; { double_cl = x; }

void init_qgroups()
{
  if (CREATE == 4) system("mv outputsqs outputsqs~");

  m = QGROUP - INCOMPLETE;
  n = QGROUP;
  n1 = n-1;
  n_sqt = QGROUP * QGROUP;
  n_cube = QGROUP * n_sqt;
  n_four = QGROUP*n_cube+1;

  if (CYCLIC) init_cgroups();
  else {
    Max_atom = n_cube;
    /*if (PREP == 1) Max_atom = P3(QGROUP-2, QGROUP-1, QGROUP-1, QGROUP-1);*/
    if (QUEEN == 25 || QUEEN == 26 || PIGEON == 2) Max_atom = 2*n_cube;
    if (TRACE > 7) print_qg_encoding();
  }

  if (QUEEN == 300) { 
    set_double_cl(2); 
    SQUARE2 = Max_atom;  
    Max_atom += Max_atom+Max_atom;
  } else if (20 <= QUEEN && QUEEN <= 24) {
    set_double_cl(1); 
    SQUARE2 = Max_atom;
    Max_atom += Max_atom;
    if (PIGEON == 2 && QUEEN < 23) Max_atom += n_cube;
  } 
  Max_clause = Max_atom*Max_atom;
  if (QUEEN < 3) Max_clause *= n_sqt;
}

void print_qg_encoding()
{
  int x, y, z;

  for (x = 0; x < n; x++)
    for (y = 0; y < n; y++) 
      for (z = 0; z < n; z++) 
	printf("TRIPLE(%d, %d, %d) = %d\n", x, y, z, TRIPLE2KEY(x, y, z));
  printf("\n");
}

/**** The following function generates the input clauses for
  quaisgroup problems. ****/

int qgroup_clauses ()
     /* return 1 if an empty clause is found; otherwise, 0 */
{
  SATOINT cl_arr [ MAX_ATOM ], sign_arr [ MAX_ATOM ];
  SATOINT x, y, u, v, h;

  init_same_hole();
  if (RAMSEY > 0) locate_holes();

  /*if (qgroup_extra_unit_clauses()) return 1;*/
  if (qgroup_unit_clauses()) return 1;

  if (50 <= FORMAT && FORMAT <= 90) cyclic_cls( cl_arr, sign_arr );

  for (x = n1; x >= 0; x--) 
    for (y = n1; y >= 0; y--) if (is_same_hole(x, y) == 0)
      for (u = n1; u >= 0; u--) 
	if (is_same_hole(u, x) == 0 && is_same_hole(u, y) == 0)
	for (v = u-1; v >= 0; v--) 
	if (is_same_hole(v, x) == 0 && is_same_hole(v, y) == 0) {

	    /* x * y = u and x * y = v imply u = v. */
	    cl_arr[0] = TRIPLE2KEY(x, y, u);
	    sign_arr[cl_arr[0]] = 0;
	    cl_arr[1] = TRIPLE2KEY(x, y, v);
	    sign_arr[cl_arr[1]] = 0;
	    if (qg_insert_clause( cl_arr, sign_arr, 2) == 1)
	      return 1;

	    /* x * u = y and x * v = y imply u = v. */
	    cl_arr[0] = TRIPLE2KEY(x, u, y);
	    sign_arr[cl_arr[0]] = 0;
	    cl_arr[1] = TRIPLE2KEY(x, v, y);
	    sign_arr[cl_arr[1]] = 0;
	    if (qg_insert_clause( cl_arr, sign_arr, 2) == 1)
	      return 1;
	  
	    /* u * x = y and v * x = y imply u = v. */
	    cl_arr[0] = TRIPLE2KEY(u, x, y);
	    sign_arr[cl_arr[0]] = 0;
	    cl_arr[1] = TRIPLE2KEY(v, x, y);
	    sign_arr[cl_arr[1]] = 0;
	    if (qg_insert_clause( cl_arr, sign_arr, 2) == 1)
	      return 1;
	  }


  if (QUEEN == 0) { if (qg0(cl_arr, sign_arr)) return 1; }
  else if (QUEEN == 1) { if (qg1(cl_arr, sign_arr)) return 1; }
  else if (QUEEN == 2) { if (qg2(cl_arr, sign_arr)) return 1; }
  else if (QUEEN == 3) { if (qg3(cl_arr, sign_arr)) return 1; }
  else if (QUEEN == 4) { if (qg4(cl_arr, sign_arr)) return 1; }
  else if (QUEEN == 5) { if (qg5(cl_arr, sign_arr)) return 1; }
  else if (QUEEN == 6) { if (qg6(cl_arr, sign_arr)) return 1; }
  else if (QUEEN == 7) { if (qg7(cl_arr, sign_arr)) return 1; }
  else if (QUEEN == 8) { if (qg8(cl_arr, sign_arr)) return 1; }
  else if (QUEEN == 9) { if (qg9(cl_arr, sign_arr)) return 1; }
  else if (QUEEN == 10) { if (qg11(cl_arr, sign_arr)) return 1;
			  if (qg12(cl_arr, sign_arr)) return 1; }
  else if (QUEEN == 11) { if (qg11(cl_arr, sign_arr)) return 1; } 
  else if (QUEEN == 12) { if (qg12(cl_arr, sign_arr)) return 1; } 
  else if (QUEEN == 13) { if (qg13(cl_arr, sign_arr)) return 1; } 
  else if (QUEEN == 14) { if (qg14(cl_arr, sign_arr)) return 1; }
  else if (QUEEN == 15) { if (qg15(cl_arr, sign_arr)) return 1; }
  else if (QUEEN == 16) { if (qg16(cl_arr, sign_arr)) return 1; }
  else if (QUEEN == 18) { if (qg0(cl_arr, sign_arr)) return 1; 
			  if (diagonal_cls(cl_arr, sign_arr)) return 1; }
  else if (QUEEN == 19) { if (qg1(cl_arr, sign_arr)) return 1; 
			  if (diagonal_cls321(cl_arr, sign_arr)) return 1; }
  else if (QUEEN == 20) { set_double_cl(0);
			  if (qg0(cl_arr, sign_arr, 1)) return 1; 
			  if (orthogonal_cls(1, cl_arr, sign_arr, 1)) return 1;
			  set_double_cl(1);
			}
  else if (QUEEN == 21) { set_double_cl(0);
			  if (qg1(cl_arr, sign_arr, 1)) return 1; 
			  if (orthogonal_cls(1, cl_arr, sign_arr, 2)) return 1;
			  set_double_cl(1);
			}
  else if (QUEEN == 22) { set_double_cl(0);
			  if (qg2(cl_arr, sign_arr, 1)) return 1; 
			  if (orthogonal_cls(1, cl_arr, sign_arr, 3)) return 1;
			  set_double_cl(1);
			}
  else if (QUEEN == 23) { if (qg13(cl_arr, sign_arr)) return 1; }
  else if (QUEEN == 24) { set_double_cl(0);
			  if (symmetric_cls(SQUARE2, cl_arr, sign_arr)) return 1;
			  if (qg0(cl_arr, sign_arr, 1)) return 1; 
			  set_double_cl(1);
			}
  else if (QUEEN == 25) { if (qg25(cl_arr, sign_arr)) return 1; }
  else if (QUEEN == 26) { if (qg26(cl_arr, sign_arr)) return 1; }
  else if (QUEEN == 30) { if (qg13(cl_arr, sign_arr)) return 1;
			  if (qg0(cl_arr, sign_arr)) return 1; }
  else if (QUEEN == 31) { if (qg13(cl_arr, sign_arr)) return 1;
			  if (qg1(cl_arr, sign_arr)) return 1; }
  else if (QUEEN == 32) { if (qg13(cl_arr, sign_arr)) return 1;
			  if (qg2(cl_arr, sign_arr)) return 1; }
  else if (QUEEN == 33) { if (qg3(cl_arr, sign_arr)) return 1; 
			  if (diagonal_cls(cl_arr, sign_arr)) return 1; }
  else if (QUEEN == 34) { if (qg30(4, cl_arr, sign_arr)) return 1; }
  else if (QUEEN == 35) { if (qg30(5, cl_arr, sign_arr)) return 1; }
  else if (QUEEN == 36) { if (qg30(1, cl_arr, sign_arr)) return 1; }
  else if (QUEEN == 39) { if (qg3(cl_arr, sign_arr)) return 1; 
			  if (qg12(cl_arr, sign_arr)) return 1; }
  else if (QUEEN == 42) { if (qg4(cl_arr, sign_arr)) return 1; 
			  if (qg12(cl_arr, sign_arr)) return 1; }
  else if (QUEEN == 44) { if (qg4(cl_arr, sign_arr)) return 1; 
			  if (diagonal_cls(cl_arr, sign_arr)) return 1; }
  else if (QUEEN == 60) { if (qg60(cl_arr, sign_arr)) return 1; }
  else if (QUEEN == 61) { if (qg0(cl_arr, sign_arr)) return 1; 
			  if (symm_trasversal(cl_arr, sign_arr)) return 1; } 
  else if (QUEEN == 71) { if (qg71(cl_arr, sign_arr)) return 1; }
  else if (QUEEN == 81) { if (qg81(cl_arr, sign_arr)) return 1; }
  else if (QUEEN == 95) { if (qg0(cl_arr, sign_arr)) return 1;
			  if (qg12(cl_arr, sign_arr)) return 1; }
  else if (QUEEN == 96) { if (qg00(cl_arr, sign_arr, 1)) return 1; }
  else if (QUEEN == 97) { if (qg00(cl_arr, sign_arr, 1)) return 1; 
			  if (qg00(cl_arr, sign_arr, 1)) return 1; }
  else if (QUEEN == 98) { if (qg9(cl_arr, sign_arr)) return 1;
			  if (diagonal_cls(cl_arr, sign_arr)) return 1; }
  else if (QUEEN == 99) { if (qg9(cl_arr, sign_arr)) return 1;
			  if (diagonal_cls321(cl_arr, sign_arr)) return 1; }
  else if (QUEEN == 101) { if (qg101(cl_arr, sign_arr)) return 1; } 
  else if (QUEEN == 102) { if (qg102(cl_arr, sign_arr)) return 1; } 
  else if (QUEEN == 104) { if (qg104(cl_arr, sign_arr)) return 1; } 
  else if (QUEEN == 105) { if (qg00(cl_arr, sign_arr, 1)) return 1; 
			   transpose_square2();
			   if (qg00(cl_arr, sign_arr, 0)) return 1; }
  else if (QUEEN == 106) { if (qg00(cl_arr, sign_arr, 1)) return 1; 
			   conjugate321_square();
			   if (qg00(cl_arr, sign_arr, 0)) return 1; }
  else if (QUEEN > 110) { if (co_cycle_cls(cl_arr, sign_arr)) return 1; }

  if (multi_squares() == 1) { 
    set_double_cl(0); 
    if (orthogonal_cls(1, cl_arr, sign_arr, 0)) return 1; 
    set_double_cl(1);
  } else if (multi_squares() == 2) {
    set_double_cl(0); 
    if (orthogonal_cls(3, cl_arr, sign_arr, 0)) return 1; 
    set_double_cl(2);
  }

  /*  * is closed.
  (i*j=1) v ... v (i*j=n)     (1<=i,j<=n)  value of product
  (i*1=j) v ... v (i*n=j)     (1<=i,j<=n)  right solution
  (1*i=j) v ... v (n*i=j)     (1<=i,j<=n)  left solution
  */
  for (x = n1; x >= 0; x--)
    for (y = n1; y >= 0; y--) if (is_same_hole(x, y) == 0) {
      h = 0;
      for (u = n1; u >= 0; u--) 
	if (is_same_hole(u, x) == 0 && is_same_hole(u, y) == 0) {
	  cl_arr[h] = TRIPLE2KEY(u, x, y);
	  sign_arr[cl_arr[h++]] = 1;
	}
      if (qg_insert_clause( cl_arr, sign_arr, h) == 1)	return 1;

      h = 0;
      for (u = n1; u >= 0; u--) 
	if (is_same_hole(u, x) == 0 && is_same_hole(u, y) == 0) {
	  cl_arr[h] = TRIPLE2KEY(x, u, y);
	  sign_arr[cl_arr[h++]] = 1;
	}
      if (qg_insert_clause( cl_arr, sign_arr, h) == 1)	return 1;

      h = 0;
      for (u = n1; u >= 0; u--) 
	if (is_same_hole(u, x) == 0 && is_same_hole(u, y) == 0) {
	  cl_arr[h] = TRIPLE2KEY(x, y, u);
	  sign_arr[cl_arr[h++]] = 1;
	}
      if (qg_insert_clause( cl_arr, sign_arr, h) == 1)	return 1;
    }

  return 0;
}


void locate_holes () 
{
  int i, j, v, u, x, num, step[3];

  NHOLE = (n-INCOMPLETE) / RAMSEY;

  /*
  if (num_of_hole && (NHOLE > num_of_hole)) {
    NHOLE = num_of_hole;
  }*/

  num_of_hole = RAMSEY * NHOLE;
  printf("%d holes of size %d for (%d-%d)\n\n", NHOLE, RAMSEY, n, INCOMPLETE);

  step[0] = 0; step[1] = SQUARE2; step[2] = SQUARE2+SQUARE2;
  num = 1 + multi_squares();
 
  for (i = 0; i < NHOLE; i++) 
    for (u = i; u < num_of_hole; u += NHOLE)
      for (v = i; v < num_of_hole; v += NHOLE) {
	set_same_hole(u, v);
	for (x = 0; x < n; x++) {
	  if (x >= m || (x % NHOLE != i)) {
	    insert_cl_1(TRIPLE2KEY(u, v, x), FF);
	    insert_cl_1(TRIPLE2KEY(u, x, v), FF);
	    insert_cl_1(TRIPLE2KEY(x, u, v), FF);
	  } else {
	    insert_cl_1(TRIPLE2KEY(u, v, x), TT);
	    insert_cl_1(TRIPLE2KEY(u, x, v), TT);
	    insert_cl_1(TRIPLE2KEY(x, u, v), TT);
	  }
	}
      }
 
  for (u = m; u < n; u++)
    for (v = m; v < n; v++)
      for (x = m; x < n; x++)
	insert_cl_1(TRIPLE2KEY(u, v, x), TT);
}

qgroup_unit_clauses ()
{
  int i, x, y;
 
  if (LINE >= 4000) qg_read_square();

  if (IDEMPOTENT) { /* x * x = x; */
    if (QUEEN != 103 || QGROUP % 2) {
      if (OUTPUT < 2) 
	if (INCOMPLETE)
	  printf("Looking for idempotent IQG%d(%d, %d).\n\n",
		 QUEEN, m, INCOMPLETE);
	else
	  printf("Looking for idempotent QG%d(%d).\n\n", QUEEN, m);

      for (x = 0; x < m; x++) {
	if (insert_cl_1(i = TRIPLE2KEY(x, x, x), TT)) return 1;
	if ((QUEEN != 24 || (QGROUP % 2)) && double_cl) {
	  if (insert_cl_1(i+SQUARE2, TT)) return 1;
	  if (double_cl > 1 && insert_cl_1(i+SQUARE2+SQUARE2, TT)) return 1;
	}
      }
    } 
  } 
  
  if (FORMAT <= 10 || 100 <= FORMAT) {
    int step = 1;
    
    if (FORMAT >= 300) { step = 2; FORMAT -= 200; }
    else if (RAMSEY > 1) step = 2;
    else if (FORMAT > 1 && FORMAT < 10) step = FORMAT;

    /* additional clauses, i.e., a good heuristic.
       FC1: x < y-step implies x * 1 != y.  FORMAT=100
       FC2: x < y-step implies y * 1 != x.  FORMAT=200
       FR1: x < y-step implies 1 * x != y.  FORMAT=101
       FR2: x < y-step implies 1 * y != x.  FORMAT=201
       FV1: x < y-step implies x * y != 1.  FORMAT=102
       FV2: x < y-step implies y * x != 1.  FORMAT=202
       LC1: x < y-step implies x * n != y.  FORMAT=103
       LC2: x < y-step implies y * n != x.  FORMAT=203 -- by McCune
       LR1: x < y-step implies n * x != y.  FORMAT=104
       LR2: x < y-step implies n * y != x.  FORMAT=204
       LV1: x < y-step implies x * y != n.  FORMAT=105
       LV2: x < y-step implies y * x != n.  FORMAT=205
       */
    for (y = 2; y < m; y++) if (is_same_hole(n1, y) == 0)
      for (x = 0; x < y-step; x++) 
	if (is_same_hole(x, y) == 0 && is_same_hole(n1, x) == 0 &&
	    is_same_hole(x, 0) == 0 && is_same_hole(y, 0) == 0) {
	  if (FORMAT < 11) {
	    /* choose the best for each problem */
	    if (RAMSEY > 1) i = TRIPLE2KEY(x, 0, y);
	    else if (QUEEN == 2 || QUEEN == 6) i = TRIPLE2KEY(x, y, 0);
	    else if (QUEEN == 3 || QUEEN == 4) i = TRIPLE2KEY(x, y, n1);
	    else i = TRIPLE2KEY(x, n1, y);
	  }
	  else if (FORMAT == 100) i = TRIPLE2KEY(x, 0, y);
	  else if (FORMAT == 200) i = TRIPLE2KEY(y, 0, x);
	  else if (FORMAT == 101) i = TRIPLE2KEY(0, x, y);
	  else if (FORMAT == 201) i = TRIPLE2KEY(0, y, x);
	  else if (FORMAT == 102) i = TRIPLE2KEY(x, y, 0);
	  else if (FORMAT == 202) i = TRIPLE2KEY(y, x, 0);
	  else if (FORMAT == 103) i = TRIPLE2KEY(x, n1, y);
	  else if (FORMAT == 203) i = TRIPLE2KEY(y, n1, x);
	  else if (FORMAT == 104) i = TRIPLE2KEY(n1, x, y);
	  else if (FORMAT == 204) i = TRIPLE2KEY(n1, y, x);
	  else if (FORMAT == 105) i = TRIPLE2KEY(x, y, n1);
	  else if (FORMAT == 205) i = TRIPLE2KEY(y, x, n1);
	  else i = 2;
	  
	  if (insert_cl_1(i, FF)) return 1;
	}
    
    if (IDEMPOTENT && FORMAT == 10) {
      for (x = m/2; x < m-2; x++) 
	if (insert_cl_1(TRIPLE2KEY(x, n1, x+1), TT)) return 1;
    }
  }
  
  return 0;
}

qg0 (cl_arr, sign_arr)
     int cl_arr [], sign_arr [];
     /* x * y = u, z * w = u, y * x = v, w * z = v -> x = z, y = w */
{
  int x, y, z, u, v, w;
  
  if (PIGEON == 2) return (qg0_P2(cl_arr, sign_arr));
  
  for (x = n1; x > 0; x--)
    for (y = (x >= m)? m-1 : n1; y >= 0; y--) if (x != y) {
      for (z = (x < m && y < m)? n1 : m-1; z >= 0; z--) 
	if (z != x && z != y &&
	    insert_cl_all_3(TRIPLE2KEY(y, y, y),
			    TRIPLE2KEY(x, z, y), 
			    TRIPLE2KEY(z, x, y), 
			    0, cl_arr, sign_arr))
	  return 1;
      
      for (z = x-1; z >= 0; z--) 
	for (w = (z >= m)? m-1 : n1; w >= 0; w--) if (z != w) 
	  for (u = (x < m && y < m && z < m && w < m)? n1 : m-1;
	       u >= 0; u--)
	    for (v = (x < m && y < m && z < m && w < m)? n1 : m-1;
		 v >= 0; v--) 
	      if (insert_cl_4(TRIPLE2KEY(x, y, u),
			      TRIPLE2KEY(z, w, u), 
			      TRIPLE2KEY(y, x, v), 
			      TRIPLE2KEY(w, z, v), 
			      0, cl_arr, sign_arr))
		return 1;
    }
  
  return 0;
}

qg0_P2 (cl_arr, sign_arr)
     int cl_arr [], sign_arr [];
     /* x * y = u, z * w = u, y * x = v, w * z = v -> x = z, y = w */
{
  int x, y, z, w;
  
  if (omega_cls(cl_arr, sign_arr)) return 1;

  for (x = n1; x >= 0; x--)
    for (y = n1; y >= 0; y--) if (is_same_hole(x, y) == 0)
      for (z = n1; z >= 0; z--) 
	for (w = n1; w >= 0; w--) {

	  if (insert_cl_all_3(TRIPLE2KEY(x, y, w),
			      TRIPLE2KEY(y, x, z), 
			      OMEGA(x, w, z), 
			      1, cl_arr, sign_arr))
	    return 1;
	}

  return 0;
}

qg10 (cl_arr, sign_arr)
     int cl_arr [], sign_arr [];
     /* x * y = u, z * w = u, y * x = v, w * z = v -> x = z, y = w */
{
  int x, y, z, u, v, w;

  if (qg0(cl_arr, sign_arr)) return 1;

  for (x = n1; x > 0; x--)
    for (y = (x >= m)? m-1 : n1; y >= 0; y--) 
      for (z = x-1; z >= 0; z--) 
	for (w = (z >= m)? m-1 : n1; w >= 0; w--) if (y != w)
	  for (u = (x < m && y < m && z < m && w < m)? n1 : m-1;
	       u >= 0; u--) 
	    for (v = (x < m && y < m && z < m && w < m)? n1 : m-1;
		 v >= 0; v--) 
	      if (insert_cl_4(TRIPLE2KEY(x, y, u),
			      TRIPLE2KEY(z, w, u), 
			      TRIPLE2KEY(neg(y), neg(x), v), 
			      TRIPLE2KEY(neg(w), neg(z), v), 
			      0, cl_arr, sign_arr))
		return 1;

  return 0;
}

int qg1 (cl_arr, sign_arr)
     int cl_arr [], sign_arr [];
     /* x * y = u, z * w = u, v * y = x, v * w = z -> x = z, y = w */
{
  int x, y, z, u, v, w;
 
  if (PIGEON == 2) return (qg1_P2(cl_arr, sign_arr));
  if (PREP == 1) P2_vs_P3(cl_arr, sign_arr);

  for (x = n1; x >= 0; x--)
    for (y = (x >= m)? m-1 : n1; y >= 0; y--) if (x != y) {

      for (z = (x < m && y < m)? n1 : m-1; z >= 0; z--) 
	if (z != x && z != y)
	  if (RAMSEY) {
	    for (u = n1; u >= 0; u--) 
	      if (is_same_hole(y, u) &&
		  insert_cl_2(TRIPLE2KEY(x, z, u), 
			      TRIPLE2KEY(y, z, x), 
			      0, cl_arr, sign_arr))
		return 1; 
	  } else {
	    if (insert_cl_all_3(TRIPLE2KEY(y, y, y),
				TRIPLE2KEY(x, z, y), 
				TRIPLE2KEY(y, z, x), 
				0, cl_arr, sign_arr))
	      return 1; 
	  }
      
      for (z = x-1; z >= 0; z--) 
	for (w = (z >= m)? m-1 : n1; w >= 0; w--) if (y != w && z != w) 
	  for (u = (x < m && y < m && z < m && w < m)? n1 : m-1;
	       u >= 0; u--) if (u != x && u != y && u != z && u != w)
	  if (PREP == 1) {

	    if (insert_cl_all_3(TRIPLE2KEY(x, y, u),
				TRIPLE2KEY(z, w, u),
				P2(z, x, y, w), 
				TT, cl_arr, sign_arr))
	      return 1;
	    if (insert_cl_all_3(TRIPLE2KEY(u, y, x),
				TRIPLE2KEY(u, w, z),
				P3(z, x, y, w), 
				TT, cl_arr, sign_arr))
	      return 1;
	  } else {
	    for (v = (x < m && y < m && z < m && w < m)? n1 : m-1;
		 v >= 0; v--) 
	      if (insert_cl_4(TRIPLE2KEY(x, y, u),
			      TRIPLE2KEY(z, w, u),
			      TRIPLE2KEY(v, y, x), 
			      TRIPLE2KEY(v, w, z), 
			      0, cl_arr, sign_arr))
		return 1;
	  }
    }
  return 0;
}

qg1_P2 (cl_arr, sign_arr)
     int cl_arr [], sign_arr [];
     /* x * y = u, z * w = u, v * y = x, v * w = z -> x = z, y = w */
{
  int x, y, z, w;

  if (omega_cls(cl_arr, sign_arr)) return 1;

  for (x = n1; x >= 0; x--)
    for (y = n1; y >= 0; y--) if (is_same_hole(x, y) == 0)
      for (z = n1; z >= 0; z--) 
	for (w = n1; w >= 0; w--) {

	  if (insert_cl_all_3(TRIPLE2KEY(x, y, w),
			      TRIPLE2KEY(z, y, x), 
			      OMEGA(x, w, z), 
			      1, cl_arr, sign_arr))
	    return 1;
	}

  return 0;
}

int qg2 (cl_arr, sign_arr)
     int cl_arr [], sign_arr []; 
     /* Looking for (231)-COLS:
        x * y = u, z * w = u, y * v = x, w * v = z -> x = z, y = w */
     /* If looking for (312)-COLS, use:
	x * y = u, z * w = u, v * x = y, v * z = w -> x = z, y = w */
{
  int x, y, z, u, v, w;

  if (PIGEON == 2) return (qg2_P2(cl_arr, sign_arr));

  for (x = n1; x >= 0; x--)
    for (y = (x >= m)? m-1 : n1; y >= 0; y--) if (x != y) {
      for (z = (x < m && y < m)? n1 : m-1; z >= 0; z--) 
	if (z != x && z != y &&
	    insert_cl_all_3(TRIPLE2KEY(y, y, y),
			    TRIPLE2KEY(x, z, y), 
			    TRIPLE2KEY(z, y, x), 
			    0, cl_arr, sign_arr))
	  return 1;

      for (z = x-1; z >= 0; z--) 
	for (w = (z >= m)? m-1 : n1; w >= 0; w--) if (y != w && z != w) 
	  for (u = (x < m && y < m && z < m && w < m)? n1 : m-1;
	       u >= 0; u--) if (u != x && u != y && u != z && u != w)
	    for (v = (x < m && y < m && z < m && w < m)? n1 : m-1;
		 v >= 0; v--) if (v != x && v != y && v != z && v != w)
	      if (insert_cl_4(TRIPLE2KEY(x, y, u),
			      TRIPLE2KEY(z, w, u),
			      TRIPLE2KEY(y, v, x),
			      TRIPLE2KEY(w, v, z),
			      0, cl_arr, sign_arr))
		return 1;
      }
  return 0;
}

qg2_P2 (cl_arr, sign_arr)
     int cl_arr [], sign_arr [];
     /* (231) x * y = u, z * w = u, y * v = x, w * v = z -> x = z, y = w */
     /* (312) x * y = u, z * w = u, v * x = y, v * z = w -> x = z, y = w */
{
  int x, y, z, w;

  if (omega_cls(cl_arr, sign_arr)) return 1;

  for (x = n1; x >= 0; x--)
    for (y = n1; y >= 0; y--) if (is_same_hole(x, y) == 0)
      for (z = n1; z >= 0; z--) 
	for (w = n1; w >= 0; w--) {

	  if (insert_cl_all_3(TRIPLE2KEY(x, y, w),
			      TRIPLE2KEY(y, z, x), 
			      OMEGA(x, w, z), 
			      1, cl_arr, sign_arr))
	    return 1;
	}

  return 0;
}

int qg3 (cl_arr, sign_arr)
     int cl_arr [], sign_arr [];
     /* (x * y) * (y * x) = x, and its two variants */
{
  int x, y, u, v;

  if (PIGEON == 1) {
    /* x * y = u, x * u = y => x = y */
    for (x = n1; x >= 0; x--) 
      for (y = (x >= m)? m-1 : n1; y >= 0; y--) if (x != y)
	for (u = (x < m && y < m)? y-1 : m-1; u >= 0; u--) {
	  cl_arr[0] = TRIPLE2KEY(x, y, u);
	  sign_arr[cl_arr[0]] = 0;
	  if (insert_one_key(TRIPLE2KEY(x, u, y), 0,
			     cl_arr, sign_arr, 1) == 0)
	    Subsume_num++;
	  else if (qg_insert_clause( cl_arr, sign_arr, 2) == 1)
	    return 1;
	}

    /* x * y = u, u * y = x => x = y */
    for (x = n1; x >= 0; x--) 
      for (y = (x >= m)? m-1 : n1; y >= 0; y--) if (x != y)
	for (u = (x < m && y < m)? x-1 : m-1; u >= 0; u--) {
	  cl_arr[0] = TRIPLE2KEY(x, y, u);
	  sign_arr[cl_arr[0]] = 0;
	  if (insert_one_key(TRIPLE2KEY(u, y, x), 0,
			     cl_arr, sign_arr, 1) == 0)
	    Subsume_num++;
	  else if (qg_insert_clause( cl_arr, sign_arr, 2) == 1)
	    return 1;
	}
  }

  /* x * y = u, y * x = v => u * v = x */
  for (x = n1; x >= 0; x--) 
    for (y = (x >= m)? m-1 : n1; y >= 0; y--) 
      for (u = (x < m && y < m)? n1 : m-1; u >= 0; u--)
	for (v = (x < m && y < m && u < m)? n1 : m-1; v >= 0; v--) 
	  if (insert_cl_all_3(TRIPLE2KEY(x, y, u),
			      TRIPLE2KEY(y, x, v), 
			      TRIPLE2KEY(u, v, x), 
			      3, cl_arr, sign_arr))
	    return 1;

  if (PIGEON == 3 && qg0(cl_arr, sign_arr)) return 1;

  return 0;
}

int qg4 (cl_arr, sign_arr)
     int cl_arr [], sign_arr [];
     /* (x * y) * (y * x) = y, and its two variants */
{
  int x, y, u, v;

  /* x * y = u, y * x = v => u * v = y */
  for (x = n1; x >= 0; x--) 
    for (y = (x >= m)? m-1 : n1; y >= 0; y--) 
      for (u = (x < m && y < m)? n1 : m-1; u >= 0; u--) 
	for (v = (x < m && y < m && u < m)? n1 : m-1; v >= 0; v--) 
	  if (insert_cl_all_3(TRIPLE2KEY(x, y, u),
			      TRIPLE2KEY(y, x, v),
			      TRIPLE2KEY(u, v, y), 
			      3, cl_arr, sign_arr))
	    return 1;

  if (PIGEON == 3 && qg0(cl_arr, sign_arr)) return 1;
  return 0;
}

int qg5 (cl_arr, sign_arr)
     int cl_arr [], sign_arr [];
     /* ((y * x) * y) * y = x, plus its two variants */
{
  int x, y, u, v;
  
  /* x * y = u, u * x = v => v * x = y */
  for (x = n1; x >= 0; x--) 
    for (y = (x >= m)? m-1 : n1; y >= 0; y--) 
      for (u = (x < m && y < m)? n1 : m-1; u >= 0; u--) 
	for (v = (x < m && y < m && u < m)? n1 : m-1; v >= 0; v--) {

	  /* u * x = y, y * u = v imply v * u = x */
	  if (insert_cl_all_3(TRIPLE2KEY(u, x, y), 
			      TRIPLE2KEY(y, u, v),
			      TRIPLE2KEY(v, u, x), 
			      3, cl_arr, sign_arr))
	    return 1;
	}
  return 0;
}

int qg6 (cl_arr, sign_arr)
     int cl_arr [], sign_arr [];
     /* (x * y) * y = x * (x * y) and its two variants */
{
  int x, y, u, v;

  /* x * y = u, x * u = v => u * y = v */
  for (x = n1; x >= 0; x--) 
    for (y = (x >= m)? m-1 : n1; y >= 0; y--) 
      for (u = (x < m && y < m)? n1 : m-1; u >= 0; u--) 
	for (v = (x < m && y < m && u < m)? n1 : m-1; v >= 0; v--) 
	  if (insert_cl_all_3(TRIPLE2KEY(x, y, u),
			      TRIPLE2KEY(u, y, v), 
			      TRIPLE2KEY(x, u, v), 
			      3, cl_arr, sign_arr))
	    return 1;
  return 0;
}

int qg7 (cl_arr, sign_arr)
     int cl_arr [], sign_arr [];
     /* (x * y) * x = y * (x * y) and its two variants */
{
  int x, y, u, v;
 
  /* x * y = u, u * x = v => y * u = v */
  for (x = n1; x >= 0; x--) 
    for (y = (x >= m)? m-1 : n1; y >= 0; y--) 
      for (u = (x < m && y < m)? n1 : m-1; u >= 0; u--) 
	for (v = (x < m && y < m && u < m)? n1 : m-1; v >= 0; v--) 
	  /* x * y = u, u * x = v imply y * u = v */
	  if (insert_cl_all_3(TRIPLE2KEY(x, y, u), 
			      TRIPLE2KEY(u, x, v),
			      TRIPLE2KEY(y, u, v), 
			      3, cl_arr, sign_arr))
	    return 1;
  return 0;
}

int qg71 (cl_arr, sign_arr)
     int cl_arr [], sign_arr [];
     /* ((x * y) * x) * y = x, and its two variants */
{
  int x, y, u, v;

  for (x = n1; x >= 0; x--) 
    for (y = (x >= m)? m-1 : n1; y >= 0; y--) 
      for (u = (x < m && y < m)? n1 : m-1; u >= 0; u--) 
	for (v = (x < m && y < m && u < m)? n1 : m-1; v >= 0; v--) 
	  /* x * y = u, u * x = v imply v * y = x */
	  if (insert_cl_all_3(TRIPLE2KEY(x, y, u), 
			      TRIPLE2KEY(u, x, v),
			      TRIPLE2KEY(v, y, x),
			      3, cl_arr, sign_arr))
	    return 1;
  return 0;
}

int qg8 (cl_arr, sign_arr)
     int cl_arr [], sign_arr [];
     /* x * (x * y) = (y * x) and its two variants */
{
  int x, y, u, v;

  for (x = n1; x >= 0; x--) 
    for (y = (x >= m)? m-1 : n1; y >= 0; y--) 
      for (u = (x < m && y < m)? n1 : m-1; u >= 0; u--) 
	for (v = (x < m && y < m && u < m)? n1 : m-1; v >= 0; v--) 
	  /* x * y = u, y * x = v => x * u = v */
	  if (insert_cl_all_3(TRIPLE2KEY(x, y, u),
			      TRIPLE2KEY(y, x, v),
			      TRIPLE2KEY(x, u, v),
			      3, cl_arr, sign_arr))
	    return 1;
  return 0;
}

int qg81 (cl_arr, sign_arr)
     int cl_arr [], sign_arr [];
     /* (x * (x * y)) * y = x and its two variants */
{
  int x, y, u, v;

  for (x = n1; x >= 0; x--) 
    for (y = (x >= m)? m-1 : n1; y >= 0; y--) 
      for (u = (x < m && y < m)? n1 : m-1; u >= 0; u--) 
	for (v = (x < m && y < m && u < m)? n1 : m-1; v >= 0; v--) 
	  /* x * y = u, x * u = v => v * y = x */
	  if (insert_cl_all_3(TRIPLE2KEY(x, y, u),
			      TRIPLE2KEY(x, u, v),
			      TRIPLE2KEY(v, y, x),
			      3, cl_arr, sign_arr))
	    return 1;
  return 0;
}

int qg9 (cl_arr, sign_arr)
     int cl_arr [], sign_arr [];
     /* ((x * y) * y) * y ==  x */
{
  int x, y, u, v;
   
  /* x * y = u, u * y = v => v * y = x */
  for (x = n1; x >= 0; x--) 
    for (y = (x >= m)? m-1 : n1; y >= 0; y--) 
      for (u = (x < m && y < m)? n1 : m-1; u >= 0; u--) 
	for (v = (x < m && y < m && u < m)? n1 : m-1; v >= 0; v--) 
	  if (insert_cl_all_3(TRIPLE2KEY(x, y, u), 
			      TRIPLE2KEY(u, y, v),
			      TRIPLE2KEY(v, y, x), 
			      1, cl_arr, sign_arr))
	    return 1;
  return 0;
}

int qg11 (cl_arr, sign_arr)
     int cl_arr [], sign_arr [];
     /* (x * y) * y = x */
{
  int x, y, z;
 
  for (x = n1; x >= 0; x--) 
    for (y = n1; y >= 0; y--) 
      for (z = ((x < m && y < m)? n1 :
		((x < m || y < m)? m-1 : -1)); 
	   z >= 0; z--) {

	/* x * y = z -> z * y = x */
	cl_arr[0] = TRIPLE2KEY(x, y, z);
	sign_arr[cl_arr[0]] = 0;
	if (insert_one_key(TRIPLE2KEY(z, y, x), 1,
			   cl_arr, sign_arr, 1) == 0)
	    Subsume_num++;
	else if (qg_insert_clause( cl_arr, sign_arr, 2) == 1)
	  return 1;
      }
  return 0;
}

int qg12 (cl_arr, sign_arr)
     int cl_arr [], sign_arr [];
     /* x * (y * x) = y * (x * y) */
{
  int x, y, u, v, z;
  
  /* x * y = u, u * x = v => v * x = y */
  for (x = n1; x >= 0; x--) 
    for (y = (x >= m)? m-1 : n1; y >= 0; y--) 
      for (u = (x < m && y < m)? n1 : m-1; u >= 0; u--) 
	for (v = (x < m && y < m && u < m)? n1 : m-1; v >= 0; v--) 
	  for (z = (x < m && y < m && u < m && v < m)? n1 : m-1; z >= 0; z--) {

	    /* x * y = v, y * x = u, x * u = z imply y * v = z */
	    if (insert_cl_4(TRIPLE2KEY(x, y, v), 
				TRIPLE2KEY(y, x, u),
				TRIPLE2KEY(x, u, z), 
				TRIPLE2KEY(y, v, z), 
				4, cl_arr, sign_arr))
	      return 1;
	  }
  return 0;
}

int qg13 (cl_arr, sign_arr)
     int cl_arr [], sign_arr [];
     /* y * (x * y) = x */
{
  int x, y, z;
 
  for (x = n1; x >= 0; x--) 
    for (y = n1; y >= 0; y--) 
      for (z = ((x < m && y < m)? n1 :
		((x < m || y < m)? m-1 : -1)); 
	   z >= 0; z--) {

	if (insert_cl_2(TRIPLE2KEY(x, y, z),
			TRIPLE2KEY(y, z, x), 
			1, cl_arr, sign_arr))
	  return 1;

	if (insert_cl_2(TRIPLE2KEY(x, y, z),
			TRIPLE2KEY(z, x, y), 
			1, cl_arr, sign_arr))
	  return 1;
      }
  return 0;
}

qg14 (cl_arr, sign_arr)
     int cl_arr [], sign_arr [];
     /* x * y = z, y * z = u, z * u = v => u * v = x */
{
  int x, y, u, v;
 
  for (x = n1; x >= 0; x--) 
    for (y = (x >= m)? m-1 : n1; y >= 0; y--) 
      for (v = (x >= m || y >= m)? m-1 : n1; v >= 0; v--) {

	if (x >= m) 
	  /* x * v = y => v * y != u for x >= m and u >= m */
	  for (u = n1; u >= m; u--) {
	    cl_arr[0] = TRIPLE2KEY(x, v, y);
	    sign_arr[cl_arr[0]] = 0;
	    cl_arr[1] = TRIPLE2KEY(v, y, u);
	    sign_arr[cl_arr[1]] = 0;
	    if (qg_insert_clause( cl_arr, sign_arr, 2) == 1)
	      return 1;
	  }

	for (u = (v >= m || y >= m || x >= m)? m-1 : n1; u >= 0; u--) {
	  if (insert_cl_all_3(TRIPLE2KEY(x, y, u),
			      TRIPLE2KEY(y, u, v),
			      TRIPLE2KEY(u, v, x),
			      1, cl_arr, sign_arr))
	    return 1;

	  if (insert_cl_all_3(TRIPLE2KEY(x, y, u),
			      TRIPLE2KEY(y, u, v),
			      TRIPLE2KEY(v, x, y),
			      1, cl_arr, sign_arr))
	    return 1;

	  if (insert_cl_all_3(TRIPLE2KEY(x, y, u),
			      TRIPLE2KEY(u, v, x),
			      TRIPLE2KEY(y, u, v),
			      1, cl_arr, sign_arr))
	    return 1;
	}	
      }
  return 0;
}

qg15 (cl_arr, sign_arr)
     int cl_arr [], sign_arr [];
     /* x * y = z, y * z = u, z * u = v => u * v = x */
{
  int x, y, z, u, v;
 
  for (x = n1; x >= 0; x--) 
    for (y = (x >= m)? m-1 : n1; y >= 0; y--) 
      for (z = (x >= m || y >= m)? m-1 : n1; z >= 0; z--) {

	if (x >= m) {
	  /* x * z = y => z * y != u for x >= m and u >= m */
	  for (u = n1; u >= m; u--) {
	    cl_arr[0] = TRIPLE2KEY(x, z, y);
	    sign_arr[cl_arr[0]] = 0;
	    cl_arr[1] = TRIPLE2KEY(z, y, u);
	    sign_arr[cl_arr[1]] = 0;
	    if (qg_insert_clause( cl_arr, sign_arr, 2) == 1)
	      return 1;
	  }

	  /* x * z = y, z * y = u => y * u != x for x >= m */
	  for (u = m-1; u >= 0; u--) {
	    if (insert_cl_all_3(TRIPLE2KEY(x, z, y),
				TRIPLE2KEY(z, y, u),
				TRIPLE2KEY(y, u, x),
				0, cl_arr, sign_arr))
	      return 1;
	  }

	  /* v * x = y, x * y = z, y * z != u for x >=m and u >= m */
	  for (v = m-1; v > y; v--)
	    for (u = n1; u >= m; u--) {
	      if (insert_cl_all_3(TRIPLE2KEY(x, y, z),
				  TRIPLE2KEY(v, x, y),
				  TRIPLE2KEY(y, z, u),
				  0, cl_arr, sign_arr))
		return 1;
	    }
	}

	for (u = (z >= m || y >= m || x >= m)? m-1 : n1; u >= 0; u--) 
	  for (v = (z >= m || u >= m || y >= m || x >= m)? m-1 : n1;
	       v >= 0; v--) {

	    if (insert_cl_4(TRIPLE2KEY(x, y, z),
			    TRIPLE2KEY(y, z, u),
			    TRIPLE2KEY(z, u, v),
			    TRIPLE2KEY(u, v, x),
			    1, cl_arr, sign_arr))
	      return 1;

	    if (insert_cl_4(TRIPLE2KEY(x, y, z),
			    TRIPLE2KEY(z, u, v),
			    TRIPLE2KEY(u, v, x),
			    TRIPLE2KEY(y, z, u),
			    1, cl_arr, sign_arr))
	      return 1;

	    if (insert_cl_4(TRIPLE2KEY(x, y, z),
			    TRIPLE2KEY(y, z, u),
			    TRIPLE2KEY(u, v, x),
			    TRIPLE2KEY(z, u, v),
			    1, cl_arr, sign_arr))
	      return 1;

	    if (insert_cl_4(TRIPLE2KEY(x, y, z),
			    TRIPLE2KEY(y, z, u),
			    TRIPLE2KEY(z, u, v),
			    TRIPLE2KEY(v, x, y),
			    1, cl_arr, sign_arr))
	      return 1;
	  }
      }
  return 0;
}

qg16 (cl_arr, sign_arr)
     int cl_arr [], sign_arr [];
     /* x * y = z, y * z = u, z * u = v, u * v = w => v * w = x */
{
  int x, y, z, u, v, w;
 
  for (x = n1; x >= 0; x--) 
    for (y = (x >= m)? m-1 : n1; y >= 0; y--) 
      for (z = (x >= m || y >= m)? m-1 : n1; z >= 0; z--) {

	if (x >= m) {
	  /* x * z = y => z * y != u for x >= m and u >= m */
	  for (u = n1; u >= m; u--) {
	    if (insert_cl_2(TRIPLE2KEY(x, z, y),
			    TRIPLE2KEY(z, y, u),
			    0, cl_arr, sign_arr))
	      return 1;
	  }

	  /* x * z = y, z * y = u => y * u != x for x >= m */
	  for (u = m-1; u >= 0; u--) {
	    if (insert_cl_all_3(TRIPLE2KEY(x, z, y),
				TRIPLE2KEY(z, y, u),
				TRIPLE2KEY(y, u, x),
				0, cl_arr, sign_arr))
	      return 1;
	  }

	  /* v * x = y, x * y = z, y * z != u for x >=m and u >= m */
	  for (v = m-1; v > y; v--)
	    for (u = n1; u >= m; u--) {
	      if (insert_cl_all_3(TRIPLE2KEY(x, y, z),
				  TRIPLE2KEY(v, x, y),
				  TRIPLE2KEY(y, z, u),
				  0, cl_arr, sign_arr))
		return 1;
	    }
	}

	if (x != y || x != z || y != z) 
	for (u = (z >= m || y >= m || x >= m)? m-1 : n1; u >= 0; u--) {

	  if (insert_cl_all_3(TRIPLE2KEY(x, y, z),
			      TRIPLE2KEY(y, z, u),
			      TRIPLE2KEY(z, u, x),
			      0, cl_arr, sign_arr))
	    return 1;

	  for (v = (z >= m || u >= m || y >= m || x >= m)? m-1 : n1;
	       v >= 0; v--) {

	    if (PIGEON && insert_cl_4(TRIPLE2KEY(x, y, z),
				      TRIPLE2KEY(y, z, u),
				      TRIPLE2KEY(z, u, v),
				      TRIPLE2KEY(u, v, x),
				      0, cl_arr, sign_arr))
	      return 1;

	    for (w = (z >= m || u >= m || y >= m || x >= m || v >= m)? m-1 : n1;
		 w >= 0; w--) {

	      if (insert_cl_5(TRIPLE2KEY(x, y, z),
			      TRIPLE2KEY(y, z, u),
			      TRIPLE2KEY(z, u, v),
			      TRIPLE2KEY(u, v, w),
			      TRIPLE2KEY(v, w, x),
			      1, cl_arr, sign_arr))
		return 1;

	      if (insert_cl_5(TRIPLE2KEY(x, y, z),
			      TRIPLE2KEY(y, z, u),
			      TRIPLE2KEY(u, v, w),
			      TRIPLE2KEY(v, w, x),
			      TRIPLE2KEY(z, u, v),
			      1, cl_arr, sign_arr))
		return 1;

	      if (PIGEON >= 3) {
	      if (insert_cl_5(TRIPLE2KEY(x, y, z),
			      TRIPLE2KEY(y, z, u),
			      TRIPLE2KEY(z, u, v),
			      TRIPLE2KEY(v, w, x),
			      TRIPLE2KEY(u, v, w),
			      1, cl_arr, sign_arr))
		return 1;

	      if (PIGEON >= 4) {
	      if (insert_cl_5(TRIPLE2KEY(x, y, z),
			      TRIPLE2KEY(y, z, u),
			      TRIPLE2KEY(z, u, v),
			      TRIPLE2KEY(u, v, w),
			      TRIPLE2KEY(w, x, y),
			      1, cl_arr, sign_arr))
		return 1;
	    }}}
	  }
	}
      }
  return 0;
}


qg25 (cl_arr, sign_arr)
     int cl_arr [], sign_arr [];
     /* x * y = z, y * z = u, z * u = v => u * v = x */
{
  int x, y, z, u, v;

  if (phi_clauses_miss(cl_arr, sign_arr)) return 1;

  for (x = n1; x >= 0; x--) 
    for (y = (x >= m)? m-1 : n1; y >= 0; y--) if (is_same_hole(x, y) == 0)
      for (z = (x >= m || y >= m)? m-1 : n1; z >= 0; z--) 
	if (is_same_hole(x, z) == 0 &&  is_same_hole(y, z) == 0)
	for (u = (x >= m || y >= m)? m-1 : n1; u >= 0; u--) 
	  if (is_same_hole(u, z) == 0 &&  is_same_hole(u, y) == 0)
	  if (insert_cl_all_3(TRIPLE2KEY(x, y, z),
			      TRIPLE2KEY(y, z, u),
			      PHI(z, u, x),
			      1, cl_arr, sign_arr))
	    return 1;
  return 0;
}

qg26 (cl_arr, sign_arr)
     int cl_arr [], sign_arr [];
     /* x * y = z, y * z = u, z * u = v, u * v = w => v * w = x */
{
  int x, y, z, u, v;
 
  if (phi_clauses_miss(cl_arr, sign_arr)) return 1;

  for (x = n1; x >= 0; x--) 
    for (y = (x >= m)? m-1 : n1; y >= 0; y--) 
      for (z = (x >= m || y >= m)? m-1 : n1; z >= 0; z--) {
	if (x != z && insert_cl_2(PHI(x, y, z),
				  PHI(z, y, x),
				  0, cl_arr, sign_arr))
	  return 1;
      }

  for (x = n1; x >= 0; x--) 
    for (y = (x >= m)? m-1 : n1; y >= 0; y--) 
      for (z = (x >= m || y >= m)? m-1 : n1; z >= 0; z--) 
	for (u = (x >= m || y >= m)? m-1 : n1; u >= 0; u--) 
	  for (v = (x >= m || y >= m || z >= m || u >= m)? m-1 : n1; 
	       v >= 0; v--) 
	    if (insert_cl_4(
			    TRIPLE2KEY(x, y, z),
			    TRIPLE2KEY(y, z, u),
			    TRIPLE2KEY(z, u, v),
			    PHI(u, v, x),
			    1, cl_arr, sign_arr))
	      return 1;
  return 0;
}

phi_clauses_miss (cl_arr, sign_arr)
     int cl_arr [], sign_arr [];
{
  int x, y, z, u;

  for (x = n1; x >= 0; x--) {
    if (insert_cl_1(PHI(x, x, x), 1)) return 1;
    
    for (y = (x >= m)? m-1 : n1; y >= 0; y--) if (x != y) {
      if (insert_cl_1(PHI(x, y, x), 0)) return 1;
      if (insert_cl_1(PHI(x, y, y), 0)) return 1;
    }
  }

  for (x = n1; x >= 0; x--) 
    for (y = (x >= m)? m-1 : n1; y >= 0; y--) if (is_same_hole(x, y) == 0) {
      for (z = (x >= m || y >= m)? m-1 : n1; z >= 0; z--) 
	if (is_same_hole(x, z) == 0 && is_same_hole(y, z) == 0) {

	if (x >= m) {
	  /* x * z = y => z * y != u for x >= m and u >= m */
	  for (u = n1; u >= m; u--) {
	    if (insert_cl_2(TRIPLE2KEY(x, y, z),
			    TRIPLE2KEY(y, z, u),
			    0, cl_arr, sign_arr))
	      return 1;
	  }
	}

	for (u = (z >= m || y >= m)? m-1 : n1; u >= 0; u--) {
	  /* define PI */
	  if (insert_cl_all_3(TRIPLE2KEY(x, y, z),
			      TRIPLE2KEY(y, z, u),
			      PHI(x, y, u),
			      1, cl_arr, sign_arr))
	    return 1;
	}
      }
    }

  for (x = n1; x >= 0; x--) 
    for (y = (x >= m)? m-1 : n1; y >= 0; y--) 
      for (z = (x >= m || y >= m)? m-1 : n1; z >= 0; z--) {

	if (x != z) 
	  for (u = (x >= m || y >= m)? m-1 : n1; u >= 0; u--) {

	    if (insert_cl_2(PHI(u, z, y),
			    PHI(u, x, y),
			    0, cl_arr, sign_arr))
	      return 1;

	    if (insert_cl_2(PHI(u, y, z),
			    PHI(u, y, x),
			    0, cl_arr, sign_arr))
	      return 1;

	    if (insert_cl_2(PHI(z, u, y),
			    PHI(x, u, y),
			    0, cl_arr, sign_arr))
	      return 1;
	  }
      }

  return 0;
}

int qg101 (cl_arr, sign_arr)
     int cl_arr [], sign_arr [];
     /* *(y, *(y, *(y, x))) ==  x */
{
  int x, y, u, v;
  
  /* y * x = u, y * u = v => y * v = x */
  for (x = n1; x >= 0; x--) 
    for (y = (x >= m)? m-1 : n1; y >= 0; y--) 
      for (u = (x < m && y < m)? n1 : m-1; u >= 0; u--) 
	for (v = (x < m && y < m && u < m)? n1 : m-1; v >= 0; v--) 
	  if (insert_cl_all_3(TRIPLE2KEY(y, x, u), 
			      TRIPLE2KEY(y, u, v),
			      TRIPLE2KEY(y, v, x), 
			      1, cl_arr, sign_arr))
	    return 1;
  return 0;
}

int qg102 (cl_arr, sign_arr)
     int cl_arr [], sign_arr [];
     /* f(f(f(f(y, x),x),y),y) == x */
{
  int x, y, z, u, v;
 
  for (x = n1; x >= 0; x--) 
    for (y = (x >= m)? m-1 : n1; y >= 0; y--) 
      for (z = (x >= m || y >= m)? m-1 : n1; z >= 0; z--) 
	for (u = (z >= m || y >= m || x >= m)? m-1 : n1; u >= 0; u--) 
	  for (v = (z >= m || u >= m || y >= m || x >= m)? m-1 : n1;
	       v >= 0; v--) 
	    /* y * x = z, z * x = u, u * y = v, v * y = x */
	    if (insert_cl_4(TRIPLE2KEY(y, x, z),
			    TRIPLE2KEY(z, x, u),
			    TRIPLE2KEY(u, y, v),
			    TRIPLE2KEY(v, y, x),
			    1, cl_arr, sign_arr))
	      return 1;
  return 0;
}

cyclic_cls (cl_arr, sign_arr)
     int cl_arr [], sign_arr [];
{
  int x, y, v, k;
  int m1 = m-1;
  
  if (FORMAT == 50) {
    for (x = m-3; x >= 0; x--) {
      insert_cl_2(TRIPLE2KEY(x, n-1, 0), 
		  TRIPLE2KEY(m-2, 0, n-1), 
		  TT, cl_arr, sign_arr);
    }
  } else if (FORMAT == 51) {
    for (x = m-3; x >= 0; x--) {
      insert_cl_2(TRIPLE2KEY(x, n-1, 0), 
		  TRIPLE2KEY(x-1, 0, n-1), 
		  TT, cl_arr, sign_arr);
    }
  } else if (FORMAT == 70) {
    /*  x * y = v => (m1-y) * (m1-x) = ((v + m1 - (x+y)) mod m) */
    for (x = m1; x >= 0; x--)
      for (y = x-1; y >= 0; y--) 
	for (v = m1; v >= 0; v--) {
	  int newv = m + m1 + v - (x+y); newv = newv % m; 
	  /* int newv = m1 - v; */
	  cl_arr[0] = TRIPLE2KEY(x, y, v);
	  sign_arr[cl_arr[0]] = 1;
	  if (insert_one_key(TRIPLE2KEY(m1-y, m1-x, newv), 0,
			     cl_arr, sign_arr, 1) &&
	      qg_insert_clause( cl_arr, sign_arr, 2))
	    return 1;
	  
	  /*
	  cl_arr[0] = TRIPLE2KEY(x, y, v);
	  sign_arr[cl_arr[0]] = 0;
	  if (insert_one_key(TRIPLE2KEY(m1-y, m1-x, m1-v), 1,
			     cl_arr, sign_arr, 1) &&
	      qg_insert_clause( cl_arr, sign_arr, 2))
	    return 1;
	    */
	}
  } else {
    /* FORMAT == 55--60:
       x * y = v => (x-k) * (y-k) = (v-k) */
    k = get_k_flag();
    if (k < 1) k = 1;

    for (v = m1; v >= k; v--)
      for (x = m1; x >= k; x--) {

	if (FORMAT >= 58) {
	  for (y = m; y < n; y++) {
	    if (insert_cl_2(TRIPLE2KEY(v, x, y),
			    TRIPLE2KEY(v-k, x-k, y),
			    TT, cl_arr, sign_arr)) return 1;
	    if (FORMAT >= 59) {
	      if (insert_cl_2(TRIPLE2KEY(v-k, x-k, y),
			      TRIPLE2KEY(v, x, y),
			    TT, cl_arr, sign_arr)) return 1;
	    }
	  }
	}

	for (y = (FORMAT == 55)? x-1 : m1; y >= k; y--) {
	  if (insert_cl_2(TRIPLE2KEY(v, x, y),
			  TRIPLE2KEY(v-k, x-k, y-k),
			  TT, cl_arr, sign_arr)) return 1;
	  
	  if (FORMAT > 56) {
	    if (insert_cl_2(TRIPLE2KEY(v-k, x-k, y-k),
			    TRIPLE2KEY(v, x, y),
			    TT, cl_arr, sign_arr)) return 1;
	  }
	}
      }
  }
  return 0;
}

int P2_vs_P3(cl_arr, sign_arr)
     int cl_arr [], sign_arr [];
     /* P2(x, y, z, w) xor P3(x, y, z, w) */
{
  int x, y, z, w;
 
  for (z = n-2; z >= 0; z--) 
    for (x = n1; x > z; x--)
      for (y = (x >= m)? m-1 : n1; y >= 0; y--) if (x != y) 
	for (w = (z >= m)? m-1 : n1; w >= 0; w--) if (y != w && z != w) {

	  /* printf("P3(%d, %d, %d, %d) = %d\n", z, x, y, w, P3(z, x, y, w));*/

	  if (insert_cl_2(P2(z, x, y, w), 
			  P3(z, x, y, w), 
			  0, cl_arr, sign_arr))
	    return 1;
        }
  return 0;
}

symmetric_cls (num, cl_arr, sign_arr)
     int num, cl_arr[], sign_arr[];
{
  int x, y, u;

  if (QGROUP % 2 == 0) {
    if (get_addk() == 1) {
      /* x * x = 0 for all x */
      printf("assuming x . x = 0\n");
      for (u = QGROUP-INCOMPLETE-1; u >= 0; u--) {
	if (insert_cl_1(x = num+TRIPLE2KEY(u, u, 0), TT)) return 1;
      }
    } else if (get_addk() == 2) {
      /* x * x = n for all x */
      printf("assuming x . x = %d\n", QGROUP-1);
      for (u = n1-INCOMPLETE; u >= 0; u--) 
	if (insert_cl_1(x = num+TRIPLE2KEY(u, u, n1), TT)) return 1;
    } else if (get_addk() == 3) {
      printf("assuming x . x = a for some a\n");
      for (u = n1-INCOMPLETE; u >= 0; u--) 
	for (x = m-1; x > 0; x--)
	  for (y = x-1; y >= 0; y--) {
	    /* there exists y for all x . x * x = y. */
	    cl_arr[0] = num+TRIPLE2KEY(x, x, u);
	    cl_arr[1] = num+TRIPLE2KEY(y, y, u);

	    Clause_num++;
	    sign_arr[cl_arr[0]] = 0;
	    sign_arr[cl_arr[1]] = 1;
	    if (insert_clause ( cl_arr, sign_arr, 2)) return 1;

	    Clause_num++;
	    sign_arr[cl_arr[0]] = 1;
	    sign_arr[cl_arr[1]] = 0;
	    if (insert_clause ( cl_arr, sign_arr, 2)) return 1;
	  }
    }
  }

  for (u = n1-INCOMPLETE; u >= 0; u--) 
    for (x = (u >= m)? m-1 : n1; x > 0; x--)
      for (y = (x >= m)? m-1 : x-1; y >= 0; y--) {
	cl_arr[0] = num+TRIPLE2KEY(x, y, u);
	cl_arr[1] = num+TRIPLE2KEY(y, x, u);
      
	Clause_num++;
	sign_arr[cl_arr[0]] = 0;
	sign_arr[cl_arr[1]] = 1;
	if (insert_clause ( cl_arr, sign_arr, 2)) return 1;

	Clause_num++;
	sign_arr[cl_arr[0]] = 1;
	sign_arr[cl_arr[1]] = 0;
	if (insert_clause ( cl_arr, sign_arr, 2)) return 1;
      }

  return 0;
}

symm_trasversal(cl_arr, sign_arr)
     int cl_arr[], sign_arr[];
{
  int u, x, y, first, second;

  /* first position */
  first = get_addk()-1;
  if (first < 0 || first > n1/2) first = 0;

  /* second position */
  second = first+1;

  for (u = 0; u < n; u++) {
    for (x = 1; x <= first; x++) 
      for (y = 0; y < x; y++) {
	if (insert_cl_2(TRIPLE2KEY(x, first-x, u), 
			TRIPLE2KEY(y, first-y, u), 
			0, cl_arr, sign_arr))
	  return 1;  
      }

    for (x = second+1; x <= n1; x++)
	for (y = second; y < x; y++) {
	  if (insert_cl_2(TRIPLE2KEY(x, n1-x+second, u), 
			  TRIPLE2KEY(y, n1-y+second, u), 
			  0, cl_arr, sign_arr))
	    return 1;
	}

    for (x = 0; x <= first; x++) 
      for (y = second; y <= n1; y++) {
	if (insert_cl_2(
			TRIPLE2KEY(y, n1-y+second, u), 
			TRIPLE2KEY(x, first-x, u), 
			0, cl_arr, sign_arr))
	  return 1;  
      }
  }
  return 0;
}

diagonal_cls (cl_arr, sign_arr)
     int cl_arr[], sign_arr[];
{
  int x, y, u;

  for (u = 0; u < n; u++) 
    for (x = n1; x > 0; x--)
      for (y = x-1; y >= 0; y--) {
	if (insert_cl_2(TRIPLE2KEY(x, n1-x, u), 
			TRIPLE2KEY(y, n1-y, u), 
			0, cl_arr, sign_arr))
	  return 1;
      }

  if (!IDEMPOTENT) 
    for (u = 0; u < n; u++) 
      for (x = n1; x > 0; x--)
	for (y = x-1; y >= 0; y--) {
	  if (insert_cl_2(TRIPLE2KEY(x, x, u), 
			  TRIPLE2KEY(y, y, u), 
			  0, cl_arr, sign_arr))
	    return 1;
	}

  return 0;
}

diagonal_cls321 (cl_arr, sign_arr)
     int cl_arr[], sign_arr[];
{
  int x, y, u;

  if (diagonal_cls(cl_arr, sign_arr)) return 1;

  for (u = 0; u < n; u++) 
    for (x = n1; x > 0; x--)
      for (y = x-1; y >= 0; y--) {
	if (insert_cl_2(TRIPLE2KEY(u, x, n1-x), 
			TRIPLE2KEY(u, y, n1-y), 
			0, cl_arr, sign_arr))
	  return 1;
      }

  if (!IDEMPOTENT) 
    for (u = 0; u < n; u++) 
      for (x = n1; x > 0; x--)
	for (y = x-1; y >= 0; y--) {
	  if (insert_cl_2(TRIPLE2KEY(u, x, x), 
			  TRIPLE2KEY(u, y, y), 
			  0, cl_arr, sign_arr))
	    return 1;
	}
  
  return 0;
}

orthogonal_cls (num, cl_arr, sign_arr, conj)
     int num, cl_arr [], sign_arr [], conj;
     /* x * y = u, z * w = u, x # y = v, z # w = v -> x = z OR y = w */
{
  int x, y, z, u, v, w, a, b;

  if (num == 1 && PIGEON == 2) {
    return (ortho_P2(cl_arr, sign_arr, conj));
  } else {
  
  if (num == 3) {
    /* no symmtry between square 2 and square 3 */
    for (y = 2; y < n; y++) 
      for (z = 1; z < y; z++) 
	if (insert_cl_2(SQUARE2+TRIPLE2KEY(0, 1, y), 
			SQUARE2+SQUARE2+TRIPLE2KEY(0, 1, z), 
			0, cl_arr, sign_arr))
	  return 1;
  }

  while (num-- > 0) {
    if (num == 1) { a = 0; b = SQUARE2+SQUARE2; }
    else if (num == 2) { a = SQUARE2; b = SQUARE2+SQUARE2; }
    else { a = 0; b = SQUARE2; }

    for (x = n1; x > 0; x--)
      for (y = (x >= m)? m-1 : n1; y >= 0; y--) if (x != y) {

	for (z = (x < m && y < m)? n1 : m-1; z >= 0; z--) 
	  if (z != x && z != y && 
	      insert_cl_4(b+TRIPLE2KEY(y, y, y),
			  b+TRIPLE2KEY(x, z, y), 
			  a+TRIPLE2KEY(y, y, y),
			  a+TRIPLE2KEY(x, z, y), 
			  0, cl_arr, sign_arr))
	    return 1;

	for (z = x-1; z >= 0; z--) 
	  for (w = (z >= m)? m-1 : n1; w >= 0; w--) if (y != w && z != w) 
	    for (u = (x < m && y < m && z < m && w < m)? n1 : m-1; u >= 0; u--) 
	      for (v = (x < m && y < m && z < m && w < m)? n1 : m-1; v >= 0; v--) 
		if (insert_cl_4(b+TRIPLE2KEY(x, y, u),
				b+TRIPLE2KEY(z, w, u), 
				a+TRIPLE2KEY(x, y, v), 
				a+TRIPLE2KEY(z, w, v), 
				0, cl_arr, sign_arr))
		  return 1;
      }
  }
  return 0;
}}

ortho_P2 (cl_arr, sign_arr, conj)
     int cl_arr [], sign_arr [], conj;
     /* x * y = u, z * w = u, x . y = v, z . w = v -> x = z, y = w */
{
  int x, y, z, w;
  int offset = (conj)? SQUARE2+n_cube : SQUARE2;

  if (phi_cls_P2(offset, cl_arr, sign_arr)) return 1;

  for (x = n1; x >= 0; x--)
    for (y = n1; y >= 0; y--) if (is_same_hole(x, y) == 0)
      for (z = n1; z >= 0; z--) 
	for (w = n1; w >= 0; w--) {

	  if (insert_cl_all_3(SQUARE2+TRIPLE2KEY(x, y, w),
			      (conj == 0)? TRIPLE2KEY(x, y, z) :
			      (conj == 1)? TRIPLE2KEY(y, x, z) :
			      (conj == 2)? TRIPLE2KEY(z, y, x) :
			      (conj == 3)? TRIPLE2KEY(y, z, x) :
			      TRIPLE2KEY(x, y, z),
			      offset+PHI(x, w, z), 
			      1, cl_arr, sign_arr))
	    return 1;
	}

  return 0;
}

int qg_insert_clause ( cl_arr, sign_arr, total )
     int cl_arr[], sign_arr[], total;
{ 
  if (double_cl) {
    int i, j, cl_arr2[100];

    if (double_cl > 1) {
      int cl_arr3[100];
      int line2 = SQUARE2+SQUARE2;

      for (i = 0; i < total; i++) {
	j = cl_arr[i];
	cl_arr3[i] = j+line2;
	sign_arr[j+line2] = sign_arr[j];
      }
      Clause_num++;
      if (insert_clause ( cl_arr3, sign_arr, total )) return 1;
    }

    for (i = 0; i < total; i++) {
      j = cl_arr[i];
      cl_arr2[i] = j+SQUARE2;
      sign_arr[j+SQUARE2] = sign_arr[j];
    }
    Clause_num++;
    if (insert_clause ( cl_arr2, sign_arr, total )) return 1;
  }

  Clause_num++;
  if (insert_clause ( cl_arr, sign_arr, total )) { return 1; }
  return 0;
}

qg_read_square ()
{
  int i, j, k;
  char name[20];

  sprintf(name, "l%d.in", LINE%1000);
  if (Input_sqs == NULL) { open_input_sqs(name); }
  if (fscanf(Input_sqs, "%d", &i) == EOF) {
    printf("No more blocks.\n"); 
    fclose(Input_sqs);
    Input_sqs = NULL;
    return 1;
  }

  printf("Read in a square:\n");
  for (i = 0; i < n; i++) {
    printf("\n  ");
    for (j = 0; j < n; j++) {
      if (fscanf(Input_sqs, "%d", &k) == EOF) { printf("\n"); return 0; }
      if (is_same_hole(i, j)) {
	if (k >= 0) { 
	  printf("Bad input: %d (-1 expected)\n", k); exit(0); 
	} 
	printf("-1 ");
      } else if (k >= 0) {
	printf("%d ", k);
	if (insert_cl_1(TRIPLE2KEY(i, j, k), TT)) bug(9);
      } else
	printf("-%d ", QGROUP);
    }
  }
  printf("\n\n");

  return 0;
}

int omega_cls (cl_arr, sign_arr)
     int cl_arr [], sign_arr [];
{
  int x, y, u, v;

  for (x = 0; x < n; x++)
    for (y = 0; y < n; y++)
      for (u = 0; u < n; u++)
	for (v = u+1; v < n; v++) {
	  if (insert_cl_2(OMEGA(v, x, y), 
			  OMEGA(u, x, y), 
			  FF, cl_arr, sign_arr)) return 1;
	}
  return 0;
}

int phi_cls_P2 (offset, cl_arr, sign_arr)
     int offset, cl_arr [], sign_arr [];
{
  int x, y, u, v;

  for (x = 0; x < n; x++)
    for (y = 0; y < n; y++)
      for (u = 0; u < n; u++)
	for (v = u+1; v < n; v++) {
	  if (insert_cl_2(offset+PHI(v, x, y), 
			  offset+PHI(u, x, y), 
			  FF, cl_arr, sign_arr)) return 1;
	}
  return 0;
}

qgroup_extra_unit_clauses ()
{
  if (insert_cl_1(TRIPLE2KEY(0, 0, 0), TT)) return 1;
  if (insert_cl_1(TRIPLE2KEY(0, 1, 2), TT)) return 1;
  if (insert_cl_1(TRIPLE2KEY(0, 2, 3), TT)) return 1;
  if (insert_cl_1(TRIPLE2KEY(0, 3, 4), TT)) return 1;
  if (insert_cl_1(TRIPLE2KEY(0, 4, 1), TT)) return 1;

  if (insert_cl_1(TRIPLE2KEY(1, 0, 3), TT)) return 1;
  if (insert_cl_1(TRIPLE2KEY(1, 1, 1), TT)) return 1;
  if (insert_cl_1(TRIPLE2KEY(1, 2, 4), TT)) return 1;
  if (insert_cl_1(TRIPLE2KEY(1, 3, 2), TT)) return 1;
  if (insert_cl_1(TRIPLE2KEY(1, 4, 0), TT)) return 1;

  if (insert_cl_1(TRIPLE2KEY(2, 0, 4), TT)) return 1;
  if (insert_cl_1(TRIPLE2KEY(2, 1, 0), TT)) return 1;
  if (insert_cl_1(TRIPLE2KEY(2, 2, 2), TT)) return 1;
  if (insert_cl_1(TRIPLE2KEY(2, 3, 1), TT)) return 1;
  if (insert_cl_1(TRIPLE2KEY(2, 4, 3), TT)) return 1;

  if (insert_cl_1(TRIPLE2KEY(3, 0, 1), TT)) return 1;
  if (insert_cl_1(TRIPLE2KEY(3, 1, 4), TT)) return 1;
  if (insert_cl_1(TRIPLE2KEY(3, 2, 0), TT)) return 1;
  if (insert_cl_1(TRIPLE2KEY(3, 3, 3), TT)) return 1;
  if (insert_cl_1(TRIPLE2KEY(3, 4, 2), TT)) return 1;

  if (insert_cl_1(TRIPLE2KEY(4, 0, 2), TT)) return 1;
  if (insert_cl_1(TRIPLE2KEY(4, 1, 3), TT)) return 1;
  if (insert_cl_1(TRIPLE2KEY(4, 2, 1), TT)) return 1;
  if (insert_cl_1(TRIPLE2KEY(4, 3, 0), TT)) return 1;
  if (insert_cl_1(TRIPLE2KEY(4, 4, 4), TT)) return 1;

  if (insert_cl_1(SQUARE2+TRIPLE2KEY(0, 0, 0), TT)) return 1;
  if (insert_cl_1(SQUARE2+TRIPLE2KEY(0, 1, 3), TT)) return 1;
  if (insert_cl_1(SQUARE2+TRIPLE2KEY(0, 2, 4), TT)) return 1;
  if (insert_cl_1(SQUARE2+TRIPLE2KEY(0, 3, 1), TT)) return 1;
  if (insert_cl_1(SQUARE2+TRIPLE2KEY(0, 4, 2), TT)) return 1;

  if (insert_cl_1(SQUARE2+TRIPLE2KEY(1, 0, 2), TT)) return 1;
  if (insert_cl_1(SQUARE2+TRIPLE2KEY(1, 1, 1), TT)) return 1;
  if (insert_cl_1(SQUARE2+TRIPLE2KEY(1, 2, 0), TT)) return 1;
  if (insert_cl_1(SQUARE2+TRIPLE2KEY(1, 3, 4), TT)) return 1;
  if (insert_cl_1(SQUARE2+TRIPLE2KEY(1, 4, 3), TT)) return 1;

  if (insert_cl_1(SQUARE2+TRIPLE2KEY(2, 0, 3), TT)) return 1;
  if (insert_cl_1(SQUARE2+TRIPLE2KEY(2, 1, 4), TT)) return 1;
  if (insert_cl_1(SQUARE2+TRIPLE2KEY(2, 2, 2), TT)) return 1;
  if (insert_cl_1(SQUARE2+TRIPLE2KEY(2, 3, 0), TT)) return 1;
  if (insert_cl_1(SQUARE2+TRIPLE2KEY(2, 4, 1), TT)) return 1;

  if (insert_cl_1(SQUARE2+TRIPLE2KEY(3, 0, 4), TT)) return 1;
  if (insert_cl_1(SQUARE2+TRIPLE2KEY(3, 1, 2), TT)) return 1;
  if (insert_cl_1(SQUARE2+TRIPLE2KEY(3, 2, 1), TT)) return 1;
  if (insert_cl_1(SQUARE2+TRIPLE2KEY(3, 3, 3), TT)) return 1;
  if (insert_cl_1(SQUARE2+TRIPLE2KEY(3, 4, 0), TT)) return 1;

  if (insert_cl_1(SQUARE2+TRIPLE2KEY(4, 0, 1), TT)) return 1;
  if (insert_cl_1(SQUARE2+TRIPLE2KEY(4, 1, 0), TT)) return 1;
  if (insert_cl_1(SQUARE2+TRIPLE2KEY(4, 2, 3), TT)) return 1;
  if (insert_cl_1(SQUARE2+TRIPLE2KEY(4, 3, 2), TT)) return 1;
  if (insert_cl_1(SQUARE2+TRIPLE2KEY(4, 4, 4), TT)) return 1;

  return 0;
}
